#include "init_tmc2590.h"

#include "grbl/hal.h"

#define TMC_SCK_PIN (27)
#define TMC_CSN_PIN (28)
#define TMC_SDI_PIN (26)

// initialize the 3 TMC2590 stepper motor controllers
// bit bang 60 bit sequences to transmit settings to all three
// of them

#define REPEAT3(x) { (x), (x), (x) }
// TODO probably store this in settings somehow
const static struct tmc2590_config {
  enum tmc2590_reg reg;
  uint32_t val[3]; // value for each axis
} tmc_config[] = {
    // blank time 24 cycles, chopper enabled, fixed toff, hysteresis dec 16 cycles,
    // no hysteresis offset, hysteresis start offset 3, off time 120 cycles
    { CHOPCONF, REPEAT3(TBL(1)|HEND(3)|HSTRT(3)|TOFF(4)) },
    // stallguard2 enabled, stallguard threshold 0, current scale 21
    { SGCSCONF, REPEAT3(SFILT(1)|CS(21)) }, // should give a peak current of 1.6 A
    // medium gate drive for both high and low side,
    // short to ground detection enabled
    // step/dir interface enabled
    // current sense resistor max voltage set to 173mV,
    // motor dampening enabled
    // Vs fault detection enabled
    { DRVCONF, REPEAT3(SLPH(3)|SLPL(3)|SLP2(0)|VSENSE(1)|EN_PFD(1)|EN_S2VS(1)) },
    // no step interpolation
    // rising edge only
    // 1/64 microstepping
    { DRVCTRL_SDOFF0, REPEAT3(MRES(2)) }
};

static uint32_t tmc2590_mask_addr(uint32_t v, enum tmc2590_reg reg) {
  uint32_t ret = v;
  switch (reg) {
  case DRVCTRL_SDOFF0:
  case DRVCTRL_SDOFF1:
    ret &= ~DRVCTRL_MASK;
    ret |= DRVCTRL_ADDR;
    break;
  case CHOPCONF:
    ret &= ~CHOPCONF_MASK;
    ret |= CHOPCONF_ADDR;
    break;
  case SMARTEN:
    ret &= ~SMARTEN_MASK;
    ret |= SMARTEN_ADDR;
    break;
  case SGCSCONF:
    ret &= ~SGCSCONF_MASK;
    ret |= SGCSCONF_ADDR;
    break;
  case DRVCONF:
    ret &= ~DRVCONF_MASK;
    ret |= DRVCONF_ADDR;
    break;
  }
  return ret;
}

static void tmc2590_csn_enable(void) {
  digitalWrite(TMC_SCK_PIN, HIGH);
  delayMicroseconds(TMC2590_TCC_US);

  digitalWrite(TMC_CSN_PIN, LOW);
  delayMicroseconds(TMC2590_CSN_DELAY_US);
}

static void tmc2590_csn_disable(void) {
  digitalWrite(TMC_SCK_PIN, HIGH);
  delayMicroseconds(TMC2590_TCH_US);

  digitalWrite(TMC_CSN_PIN, HIGH);
  delayMicroseconds(TMC2590_TCL_US);
}

static void tmc2590_send_bit(uint32_t bit) {
  digitalWrite(TMC_SCK_PIN, LOW);
  if (bit) {
    digitalWrite(TMC_SDI_PIN, HIGH);
  } else {
    digitalWrite(TMC_SDI_PIN, LOW);
  }
  delayMicroseconds(TMC2590_CLK_US);
  digitalWrite(TMC_SCK_PIN, HIGH);
  delayMicroseconds(TMC2590_CLK_US);
}

static void tmc2590_write_reg(const struct tmc2590_config* cfg) {
  uint32_t cur_reg;
  tmc2590_csn_enable();
  for (int i = 0; i < 3; i++) {
    cur_reg = cfg->val[i];
    cur_reg = tmc2590_mask_addr(cur_reg, cfg->reg);
    for (int j = 19; j >= 0; j--) {
      tmc2590_send_bit(cur_reg & (1 << j));
    }
  }
  tmc2590_csn_disable();
}

static void tmc2590_init(void) {
  pinMode(TMC_SCK_PIN, OUTPUT);
  pinMode(TMC_SDI_PIN, OUTPUT);
  pinMode(TMC_CSN_PIN, OUTPUT);

  for (int i = 0; i < sizeof(tmc_config)/sizeof(tmc_config[0]); i++) {
    tmc2590_write_reg(&tmc_config[i]);
  }
}

static driver_reset_ptr on_reset;

static void tmc2590_on_reset(void) {
  tmc2590_init();
  on_reset();
}

void my_plugin(void) {
  // TODO add message
  tmc2590_init();

  // hook into soft reset
  on_reset = hal.driver_reset;
  hal.driver_reset = tmc2590_on_reset;
}
